/***************************************************************************
*   Copyright (C) 2004 by Reto Schoelly,,,                                *
*   reto@reto-schoelly.de                                                 *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "defines.h"

#include "mahjonggboard.h"

#ifdef LINUX
#include "gamedata_path.h"
#endif

#include "tile.h"

#include "malloc.h"
#include "stdlib.h"
#include "stdio.h"
#include "time.h"
#include "math.h"
#include "string.h"

#include "tilegl.h"

// global variable for the game time needed to compute the score.
// There might be a more elegant way to solve this :-)
int game_time;
char HISCORE_FILENAME[300];


// convert x,y,z coordinates into array coordinates
int MahJonggBoard::convertCoordinates(int x, int y, int z)
{
	if((x < 0) || (x > board_width-1)
	        || (y < 0) || (y > board_height-1)
	        || (z < 0) || (z > board_depth-1))
	{
		return -1;
	}
	else
		return (z*board_width*board_height + y*board_width + x);
}


// check whether a tile is blocked or can be removed.
bool MahJonggBoard::isBlocked(int x, int y, int z)
{
	int coord;
	bool ret = true;
	bool tile_above = false;
	bool tile_left = false;
	bool tile_right = false;

	coord = convertCoordinates(x, y, z);
	if(field[coord]->getState() == TILE_LOCKED)
		return true;

	coord = convertCoordinates(x, y, z+1);
	if(coord != -1)
		if(field[coord]->getState() > TILE_EMPTY)
			tile_above = true;

	coord = convertCoordinates(x+1, y, z);
	if(coord != -1)
		if(field[coord]->getState() > TILE_EMPTY)
			tile_right = true;

	coord = convertCoordinates(x-1, y, z);
	if(coord != -1)
		if(field[coord]->getState() > TILE_EMPTY)
			tile_left = true;

	if(!tile_above && (!tile_left || !tile_right))
		ret = false;
	else
		ret = true;

	return ret;
}


bool MahJonggBoard::selectTile(int x, int y, int z)
{
	if(selected_tile_1 != -1)
		field[selected_tile_1]->setState(TILE_UNMARKED);
	if(selected_tile_2 != -1)
		field[selected_tile_2]->setState(TILE_UNMARKED);

	if(hint_pos1 != -1)
		field[hint_pos1]->setState(TILE_UNMARKED);
	if(hint_pos2 != -1)
		field[hint_pos2]->setState(TILE_UNMARKED);

	if(isBlocked(x,y,z))
	{
		selected_tile_1 = selected_tile_2 = -1;
		return false;
	}

	if(selected_tile_1 == -1)
	{
		selected_tile_1 = convertCoordinates(x,y,z);
		field[selected_tile_1]->setState(TILE_MARKED);
		return true;
	}
	else if(selected_tile_2 == -1)
	{
		selected_tile_2 = convertCoordinates(x,y,z);

		if(tilesMatch(field[selected_tile_1], field[selected_tile_2]))
		{
			if(selected_tile_1 == selected_tile_2)
			{
				field[selected_tile_1]->setState(TILE_UNMARKED);
				field[selected_tile_2]->setState(TILE_UNMARKED);
				selected_tile_1 = selected_tile_2 = -1;
				return false;
			}

			undo_list[num_turns][0] = selected_tile_1;
			undo_list[num_turns][1] = selected_tile_2;
			num_turns++;

			removeTile(selected_tile_1);
			removeTile(selected_tile_2);

			selected_tile_1 = selected_tile_2 = -1;

			return true;
		}
		else
		{
			field[selected_tile_1]->setState(TILE_UNMARKED);
			selected_tile_1 = selected_tile_2 = -1;
			return false;
		}
	}
	return false;
}


bool MahJonggBoard::undo()
{
	if(num_turns == 0)
		return false;

	num_turns--;
	num_tiles_on_board+=2;

	tileset[field[undo_list[num_turns][0]]->getType()]++;
	setTile(undo_list[num_turns][0], field[undo_list[num_turns][0]]->getType(), TILE_UNMARKED);
	tileset[field[undo_list[num_turns][1]]->getType()]++;
	setTile(undo_list[num_turns][1], field[undo_list[num_turns][1]]->getType(), TILE_UNMARKED);

	return true;
}


bool MahJonggBoard::removeTile(int num)
{
	tileset[field[num]->getType()]++;
	setTile(num, field[num]->getType(), TILE_EMPTY);
	num_tiles_on_board--;

	return true;
}


// tiles match if their types match
bool MahJonggBoard::tilesMatch(Tile *tile1, Tile *tile2)
{
	if(tile1->getType() == tile2->getType())
		return true;
	return false;
}


// find out whether a row is used and where the leftmost and rightmost tile of the row lies
bool MahJonggBoard::findBoundaryTilesInColumn(int x, int y, int z,
        int *left, int *right, char *column_used, char *column_full)
{
	int m,n,o, coord;
	int leftbound, rightbound;

	m = x;
	n = y;
	o = z;

	coord = convertCoordinates(x,y,z);
	if(field[coord] == TILE_LOCKED)
		return false;

	leftbound = 0;
	rightbound = board_width - 1;

	while(coord != -1)
	{
		if(field[coord]->getState() == TILE_LOCKED)
		{
			break;
		}

		m--;
		coord = convertCoordinates(m, n, o);
	}
	leftbound = m+1;

	coord = convertCoordinates(x,y,z);
	m = x;
	n = y;
	o = z;

	while(coord != -1)
	{
		if(field[coord]->getState() == TILE_LOCKED)
		{
			break;
		}

		m++;
		coord = convertCoordinates(m, n, o);
	}
	rightbound = m-1;

	if(leftbound > rightbound)
	{
		return false;
	}

	*column_used = 0;

	*left = *right = -1;

	for(m = leftbound; m <= rightbound; m++)
	{
		coord = convertCoordinates(m, n, o);

		if((field[coord]->getState() == TILE_UNMARKED) || (field[coord]->getState() == TILE_MARKED))
		{
			*left = m;
			*column_used = 1;
			break;
		}
	}

	for(m = rightbound; m >= leftbound; m--)
	{
		coord = convertCoordinates(m, n, o);

		if((field[coord]->getState() == TILE_UNMARKED) || (field[coord]->getState() == TILE_MARKED))
		{
			*right = m;
			*column_used = 1;
			break;
		}
	}

	if((*right == rightbound) && (*left == leftbound))
		*column_full = 1;
	else
		*column_full = 0;

	return true;
}


void MahJonggBoard::setPositionUsable(int pos)
{
	usable_positions[num_usable_positions] = pos;
	num_usable_positions++;
}


void MahJonggBoard::setPositionUnusable(int x, int y, int z)
{
	int m;
	int coord;

	coord = convertCoordinates(x,y,z);

	for(m = 0; m < num_usable_positions; m++)
	{
		if(usable_positions[m] == coord)
		{
			usable_positions[m] = -1;
			break;
		}
	}
}


// check whether a tile can be placed when generating a solvable board
bool MahJonggBoard::usabilityTest(int x, int y, int z)
{
	int coord;

	coord = convertCoordinates(x,y,z);
	if(field[coord]->getState() != TILE_UNMARKED)
		return false;
	if(!isBlocked(x,y,z))
		return true;

	return false;
}


// check the tile whether it is usable for generating a solvable board, special tiles block more fields.
// check "turtle.layout" for more information
bool MahJonggBoard::isUsable(int x, int y, int z)
{
	int special;
	bool usable[4];

	usable[0] = usable[1] = usable[2] = usable[3] = true;

	field[convertCoordinates(x,y,z)]->getSpecial(&special);

	switch(special)
	{
	case 2:
		usable[1] = usabilityTest(x,y+1,z);
		break;
	case 3:
		usable[1] = usabilityTest(x,y-1,z);
		break;
	case 4:
		usable[1] = usabilityTest(x+1,y,z);
		usable[2] = usabilityTest(x+1,y+1,z);
		usable[3] = usabilityTest(x,y+1,z);
		break;
	case 5:
		usable[1] = usabilityTest(x-1,y,z);
		usable[2] = usabilityTest(x-1,y+1,z);
		usable[3] = usabilityTest(x,y+1,z);
		break;
	case 6:
		usable[1] = usabilityTest(x,y-1,z);
		usable[2] = usabilityTest(x+1,y-1,z);
		usable[3] = usabilityTest(x+1,y,z);
		break;
	case 7:
		usable[1] = usabilityTest(x-1,y,z);
		usable[2] = usabilityTest(x-1,y-1,z);
		usable[3] = usabilityTest(x,y-1,z);
		break;
	}

	usable[0] = usabilityTest(x,y,z);

	return usable[0] && usable[1] && usable[2] && usable[3];
}


// generate a list of board positions usable for generating a solvable board
bool MahJonggBoard::generateUsablePositions()
{
	int n;
	int x,y,z;
	int coord;
	int special;

	num_usable_positions = 0;

	for(n = 0; n < BOARD_SIZE; n++)
	{
		usable_positions[n] = -1;
	}

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				coord = convertCoordinates(x,y,z);

				if(field[coord]->getState() != TILE_UNMARKED)
					continue;
				field[coord]->getSpecial(&special);
				if((special != -1) && (special != 2) && (special != 4))
					continue;

				if(isUsable(x,y,z))
					setPositionUsable(coord);
			}

	return true;
}



// place a tile on the board
int MahJonggBoard::setTile(int pos, int tile_type, int state)
{
	//	if(tileset[tile_type] == 0)
	//	if(state == TILE_UNMARKED) return 0;

	int special;

	field[pos]->getSpecial(&special);

	switch(special)
	{
	case 2:
		field[pos+board_width]->setState(state);
		field[pos+board_width]->setType(tile_type);
		break;
	case 3:
		field[pos-board_width]->setState(state);
		field[pos-board_width]->setType(tile_type);
		break;
	case 4:
		field[pos+board_width]->setState(state);
		field[pos+board_width]->setType(tile_type);
		field[pos+1]->setState(state);
		field[pos+1]->setType(tile_type);
		field[pos+1+board_width]->setState(state);
		field[pos+1+board_width]->setType(tile_type);
		break;
	case 5:
		field[pos+board_width]->setState(state);
		field[pos+board_width]->setType(tile_type);
		field[pos-1]->setState(state);
		field[pos-1]->setType(tile_type);
		field[pos-1+board_width]->setState(state);
		field[pos-1+board_width]->setType(tile_type);
		break;
	case 6:
		field[pos-board_width]->setState(state);
		field[pos-board_width]->setType(tile_type);
		field[pos-board_width+1]->setState(state);
		field[pos-board_width+1]->setType(tile_type);
		field[pos+1]->setState(state);
		field[pos+1]->setType(tile_type);
		break;
	case 7:
		field[pos-board_width]->setState(state);
		field[pos-board_width]->setType(tile_type);
		field[pos-board_width-1]->setState(state);
		field[pos-board_width-1]->setType(tile_type);
		field[pos-1]->setState(state);
		field[pos-1]->setType(tile_type);
		break;
	}

	field[pos]->setState(state);
	field[pos]->setType(tile_type);
	tileset[tile_type]--;
	num_tiles_left--;

	return special;
}


bool MahJonggBoard::isInitialized()
{
	return initialized;
}


// find out how many pairs can be removed
int MahJonggBoard::calcNumRemovablePairs()
{
	int num = 0;
	int x, y, z;
	int coord;
	int possible_tiles[NUM_TILES];
	int type;

	numtiles = 0;
	int state;
	int special;

	for(x = 0; x < NUM_TILES; x++)
	{
		possible_tiles[x] = 0;
	}

	for(x = 0; x < BOARD_SIZE; x++)
	{
		removable_tiles_list[x] = -1;
	}

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				coord = convertCoordinates(x,y,z);
				state = field[coord]->getState();

				if((state == TILE_EMPTY) || (state == TILE_LOCKED))
					continue;
				field[coord]->getSpecial(&special);
				if((special != -1) && (special != 2) && (special != 4))
					continue;

				if(!isBlocked(x,y,z))
				{
					possible_tiles[field[coord]->getType()]++;

					removable_tiles_list[numtiles] = coord;
					numtiles++;
				}
			}

	for(x = 0; x < NUM_TILES; x++)
	{
		if(possible_tiles[x] > 1)
		{
			num++;
		}
	}

	hint_pos1 = hint_pos2 = -1;

	for(x = 0; x < BOARD_SIZE; x++)
	{
		if(removable_tiles_list[x] == -1)
			break;

		type = field[removable_tiles_list[x]]->getType();
		coord = removable_tiles_list[x];

		for(y = 0; y < BOARD_SIZE; y++)
		{
			if(x == y)
				continue;
			if(removable_tiles_list[y] == -1)
				break;

			if(field[removable_tiles_list[y]]->getType() == type)
			{
				hint_pos1 = coord;
				hint_pos2 = removable_tiles_list[y];

				return num;
			}
		}
	}

	return num;
}


int MahJonggBoard::getNumTilesOnBoard()
{
	return num_tiles_on_board;
}


// choose a tile out of the tileset
int MahJonggBoard::generateTileType()
{
	int rnum;

	rnum = (int)((float)rand()/(float)RAND_MAX * (float)(NUM_TILES));

	if(rnum > (NUM_TILES-1))
	{
		rnum = 0;
	}

	while(tileset[rnum] == 0)
	{
		rnum++;
		if(rnum > (NUM_TILES-1))
		{
			rnum = 0;
		}
	}

	if(rnum == NUM_TILES)
	{
		return 0;
	}

	return rnum;
}


// shift tiles on a row
bool MahJonggBoard::resolveUsablePositions()
{
	int a,b,c,left,right;

	char column_used,column_full;

	int m;

	for(c = 0; c < board_depth; c++)
	{
		for(b = 0; b < board_height; b++)
		{
			for(a = 0; a < board_width; a++)
			{
				if(field[convertCoordinates(a,b,c)]->getState() == TILE_LOCKED)
					continue;

				findBoundaryTilesInColumn(a,b,c,&left,&right,&column_used,&column_full);
				if((column_used == 1) && (column_full == 0))
				{
					if(field[convertCoordinates(left-1,b,c)]->getState() == TILE_EMPTY)
					{
						for(m = left; m < right; m++)
						{
							field[m-1] = field[m];
						}
						field[right]->setState(TILE_EMPTY);
						return true;
					}
					else if(field[convertCoordinates(left+1,b,c)]->getState() == TILE_EMPTY)
					{
						for(m = right; m > left; m--)
						{
							field[m+1] = field[m];
						}
						return true;
						field[left]->setState(TILE_EMPTY);
					}
				}
			}
		}
	}
	return true;
}


bool MahJonggBoard::flipHoriz(int x, int y, int z)
{
	int special;
	int coord = convertCoordinates(x,y,z);
	if(coord == -1 )
		return false;
	field[coord]->getSpecial(&special);
	if(special != -1)
		return false;
	int coord2 = convertCoordinates(x+1,y,z);
	if(coord2 == -1 )
		return false;
	field[coord2]->getSpecial(&special);
	if(special != -1)
		return false;

	Tile *tile_temp;

	if(selected_tile_1 != -1)
		if(field[selected_tile_1]->getState() == TILE_MARKED)
			field[selected_tile_1]->setState(TILE_UNMARKED);
	if(selected_tile_2 != -1)
		if(field[selected_tile_2]->getState() == TILE_MARKED)
			field[selected_tile_2]->setState(TILE_UNMARKED);
	selected_tile_1 = -1;
	selected_tile_2 = -1;

	if((field[coord]->getState() == TILE_UNMARKED) && (field[coord2]->getState() == TILE_UNMARKED))
	{
		tile_temp = field[coord];
		field[coord] = field[coord2];
		field[coord2] = tile_temp;
		return true;
	}
	return false;
}

bool MahJonggBoard::flipVert(int x, int y, int z)
{
	int special;
	int coord = convertCoordinates(x,y,z);
	if(coord == -1 )
		return false;
	field[coord]->getSpecial(&special);
	if(special != -1)
		return false;
	int coord2 = convertCoordinates(x,y+1,z);
	if(coord2 == -1 )
		return false;
	field[coord2]->getSpecial(&special);
	if(special != -1)
		return false;

	Tile *tile_temp;

	if(selected_tile_1 != -1)
		if(field[selected_tile_1]->getState() == TILE_MARKED)
			field[selected_tile_1]->setState(TILE_UNMARKED);
	if(selected_tile_2 != -1)
		if(field[selected_tile_2]->getState() == TILE_MARKED)
			field[selected_tile_2]->setState(TILE_UNMARKED);
	selected_tile_1 = -1;
	selected_tile_2 = -1;

	if((field[coord]->getState() == TILE_UNMARKED) && (field[coord2]->getState() == TILE_UNMARKED))
	{
		tile_temp = field[coord];
		field[coord] = field[coord2];
		field[coord2] = tile_temp;
		return true;
	}
	return false;
}

bool MahJonggBoard::flipDepth(int x, int y, int z)
{
	int special;
	int coord = convertCoordinates(x,y,z);
	if(coord == -1 )
		return false;
	field[coord]->getSpecial(&special);
	if(special != -1)
		return false;
	int coord2 = convertCoordinates(x,y,z+1);
	if(coord2 == -1 )
		return false;
	field[coord2]->getSpecial(&special);
	if(special != -1)
		return false;

	Tile *tile_temp;

	if(selected_tile_1 != -1)
		if(field[selected_tile_1]->getState() == TILE_MARKED)
			field[selected_tile_1]->setState(TILE_UNMARKED);
	if(selected_tile_2 != -1)
		if(field[selected_tile_2]->getState() == TILE_MARKED)
			field[selected_tile_2]->setState(TILE_UNMARKED);
	selected_tile_1 = -1;
	selected_tile_2 = -1;

	if((field[coord]->getState() == TILE_UNMARKED) && (field[coord2]->getState() == TILE_UNMARKED))
	{
		tile_temp = field[coord];
		field[coord] = field[coord2];
		field[coord2] = tile_temp;
		return true;
	}
	return false;
}

bool MahJonggBoard::reorder()
{
	int x,y,z;

	num_turns = 0;

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				flipHoriz(x,y,z);
			}

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				flipDepth(x,y,z);
			}

	return true;
}


int MahJonggBoard::getRandomSeed()
{
	return random_seed;
}



int MahJonggBoard::hint()
{
	if(selected_tile_1 != -1)
		field[selected_tile_1]->setState(TILE_UNMARKED);
	if(selected_tile_2 != -1)
		field[selected_tile_2]->setState(TILE_UNMARKED);
	
	if(hint_pos1 == -1)
		return 0;
	if(hint_pos2 == -1)
		return 0;

	field[hint_pos1]->setState(TILE_MARKED);
	field[hint_pos2]->setState(TILE_MARKED);
	return 0;
}


// finally generate a solvable board
bool MahJonggBoard::generateSolvableBoard(int seed, const char *layout)
{
	int x;

	int pos = 0, pos2 = 0;
	int type;
	int special;

	num_turns = 0;

	int securitycount;

	loadLayout(layout);

	num_tiles_on_board = 0;

	if(seed == 0)
		srand(random_seed = time(NULL));
	else
		srand(random_seed = seed);

	hint_pos1 = -1;
	hint_pos2 = -1;

	for(x = 0; x < BOARD_SIZE; x++)
	{
		undo_list[x][0] = undo_list[x][1] = -1;
		if(!field[x])
			break;
		field[x]->getSpecial(&special);

		if(field[x]->getState() != TILE_EMPTY)
			continue;
		field[x]->setState(TILE_UNMARKED);
		field[x]->setType(-1);

		switch(special)
		{
		case 3:
			break;
		case 5:
			break;
		case 6:
			break;
		case 7:
			break;

		default:
			num_tiles_on_board++;
			break;
		}
	}

	num_tiles_left = num_tiles_on_board;

	securitycount = num_tiles_left +3;

	while(num_tiles_left > 0)
	{
		securitycount--;
		if(securitycount < 0)
			return false;

		generateUsablePositions();
		if(num_usable_positions <= 1)
		{
			return false;
		}

		type = generateTileType();

		pos = (int)((float)rand()/(float)RAND_MAX * (float)(num_usable_positions));
		if(pos == num_usable_positions)
		{
			pos = 0;
		}

		setTile(usable_positions[pos], type, TILE_EMPTY);

		pos2 = (int)((float)rand()/(float)RAND_MAX * (float)(num_usable_positions));
		if(pos2 == num_usable_positions)
		{
			pos2 = 0;
		}

		while(pos2 == pos)
		{
			pos2++;
			if(pos2 >= num_usable_positions)
				pos2 = 0;
		}

		setTile(usable_positions[pos2], type, TILE_EMPTY);
	}

	for(x = 0; x < BOARD_SIZE; x++)
	{
		if(!field[x])
			break;
		if(field[x]->getState() == TILE_EMPTY)
		{
			field[x]->setState(TILE_UNMARKED);
		}
	}

	initialized = true;

	return true;
}

Tile *MahJonggBoard::getTile(int x, int y, int z)
{
	int count;
	count = convertCoordinates(x,y,z);
	if(count == -1)
		return NULL;

	return field[count];
}

bool MahJonggBoard::getBoardGeometry(int *width, int *height, int *depth)
{
	if(!initialized)
		return false;

	*width = board_width;
	*height = board_height;
	*depth = board_depth;

	return true;
}


unsigned int MahJonggBoard::getDisplayList(int x, int y, int z)
{
	int coord;
	coord = convertCoordinates(x,y,z);
	if(coord == -1)
		return 0;

	if(field[coord]->getState() == TILE_UNMARKED)
	{
		return field[coord]->getTileGL()->getDisplayListUnmarked();
	}
	else if(field[coord]->getState() == TILE_MARKED)
	{
		return field[coord]->getTileGL()->getDisplayListMarked();
	}

	return 0;
}


unsigned int MahJonggBoard::getTextureName(int x, int y, int z)
{
	int coord;
	coord = convertCoordinates(x,y,z);
	if(coord == -1)
		return 0;

	return field[coord]->getTileGL()->getTextureName();
}



MahJonggBoard::MahJonggBoard()
{
	int x;

	game_time = 0;
	initialized = false;

	for(x = 0; x < BOARD_SIZE; x++)
	{
		field[x] = NULL;
	}

	for(x = 0; x < NUM_TILES; x++)
	{
		tileset[x] = 4;
	}
}


// load a layout out of a file
bool MahJonggBoard::loadLayout(const char *layout_filename)
{
	int x, y, z;
	FILE *file;
	selected_tile_1 = selected_tile_2 = -1;
	int readchar;
	int coord;
	int special;
	char string[300];
	memset(string, '\0', 300);

	char *home = NULL;

	game_time = 0;
	initialized = false;

	for(x = 0; x < BOARD_SIZE; x++)
	{
		field[x] = NULL;
	}

	for(x = 0; x < NUM_TILES; x++)
	{
		tileset[x] = 4;
	}

	strcat(string, layout_filename);

	if((file = fopen(string, "r")) == NULL)
	{
		printf("could not load layout %s", string);
		return false;
	}
#ifdef LINUX
	home = getenv("HOME");
	if(!home)
	{
		//error
	}
	else
	{
		system("mkdir -p ~/.mahjongg3d");
		memset(HISCORE_FILENAME, '\0', 300);

		strcpy(HISCORE_FILENAME, home);
		strcat(HISCORE_FILENAME, "/.mahjongg3d/");
	}

	for(x=strlen(layout_filename)-1; x > 0; x--)
	{
		if((layout_filename[x] == '\\') || (layout_filename[x] == '/'))
		{
			x++;
			break;
		}
	}

	strncat(HISCORE_FILENAME, &layout_filename[x], strlen(layout_filename) - x);
	strcat(HISCORE_FILENAME, ".hiscores");

#endif

	fscanf(file, "%d", &board_width);
	fgetc(file);
	fscanf(file, "%d", &board_height);
	fgetc(file);
	fscanf(file, "%d", &board_depth);
	fgetc(file);

	for(z = 0; z < board_depth; z++)
		for(y = 0; y < board_height; y++)
			for(x = 0; x < board_width; x++)
				field[z*board_width*board_height + y*board_width + x] = new Tile();


	for(z = 0; z < board_depth; z++)
	{
		for(y = 0; y < board_height; y++)
		{
			for(x = 0; x < board_width; x++)
			{
				coord = z*board_width*board_height + y*board_width + x;
				readchar = fgetc(file);

				switch(readchar)
				{
				case '0':
					field[coord]->setState(TILE_LOCKED);
					break;
				case '1':
					field[coord]->setState(TILE_EMPTY);
					break;
				case '2':
					special = 2;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				case '3':
					special = 3;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				case '4':
					special = 4;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				case '5':
					special = 5;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				case '6':
					special = 6;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				case '7':
					special = 7;
					field[coord]->setSpecial(special);
					field[coord]->setState(TILE_EMPTY);
					break;
				}
			}
			fgetc(file);
		}

		fgetc(file);
	}

	return true;
}


MahJonggBoard::~MahJonggBoard()
{
	int m;

	for(m = 0; m < BOARD_SIZE; m++)
	{
		if(field[m])
			delete field[m];
	}
}


