/***************************************************************************
 *   Copyright (C) 2004 by Reto Schoelly,,,                                *
 *   reto@reto-schoelly.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "defines.h"

// this directive is only needed for windows, so if
// you are compiling with linux, enter a "#define LINUX" in defines.h
#ifndef LINUX
#include "..\workspace\MahJongg3D\StdAfx.h"
#endif

#include "mahjongggl.h"
#include "tile.h"
#include "tilegl.h"
#include "mahjonggboard.h"
#include "cbitmapfile.h"

#ifdef LINUX
#include "gamedata_path.h"
#endif

#include "math.h"
#include "malloc.h"
#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#include "GL/gl.h"
#include "GL/glu.h"



// cross product calculation contributed
static inline void cross (const float a[3], const float b[3], float result[3])
{
	result[0] =  (a[1] * b[2] - a[2] * b[1]);
	result[1] = -(a[0] * b[2] - a[2] * b[0]);
	result[2] =  (a[0] * b[1] - b[1] * a[0]);
}


// dot (scalar) product calculation
static inline float dot (const float a[3], const float b[3])
{
	float result = 0;
	for (int i = 0; i < 3; i++)
		result += a[i] * b[i];
	return result;
}



MahJonggGL :: MahJonggGL()
{
	rotation_type = ROTATION_DEFAULT;
	enabled = true;

	Omni0Color[0] = 1.1f;
	Omni0Color[1] = 1.1f;
	Omni0Color[2] = 1.1f;
	Omni0Color[3] = 1.0f;

	Omni0Amb[0] = 0.6f;
	Omni0Amb[1] = 0.6f;
	Omni0Amb[2] = 0.6f;
	Omni0Amb[3] = 0.6f;

	Omni0Pos[0] = 25.0f;
	Omni0Pos[1] = 23.0f;
	Omni0Pos[2] = 15.0f;
	Omni0Pos[3] = 1.0f;

	ROT_SCALE = 0.0025f;

	int x;

	show_removable = true;

	mb_r = false;
	mouse_x = 0;
	mouse_y = 0;
	angle_x = -PI / 2.0f - 0.1f;
	angle_z = 1.0f;

	while(angle_x >= (2*PI)) // angles will not get greater 2PI and not smaller than 0
		angle_x -= 2*PI;

	while(angle_x < 0)
		angle_x += 2*PI;

	if(angle_z >= (0.49f*PI))
		angle_z = 0.49f*PI;

	while(angle_z < 0)
		angle_z = 0.0f;

	cam_x = CAMERA_DISTANCE * (float)cos(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_x;
	cam_y = CAMERA_DISTANCE * (float)sin(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_y;
	cam_z = CAMERA_DISTANCE * (float)sin(angle_z);

	float drag_here[3];
	mapToSphere (0, 0, drag_here);
	cross (drag_start, drag_here, axis);
	cosangle = dot (drag_start, drag_here);

	pick = false;

	board = NULL;

	shadow = true;

	congratulations = NULL;

	for(x = 0; x < NUM_TEXTURES; x++)
	{
		tiles[x] = NULL;
	}

	board = NULL;
	background_texture = 0;
	bg_tex = NULL;

	for (int i = 0; i < 4; i ++)
		for (int j = 0; j < 4; j ++)
			transform[4*i + j] = (i == j);
}

MahJonggGL :: ~MahJonggGL()
{
	for(int n=0; n < 10; n++)
	{
		if(number_bmp[n])
			delete number_bmp[n];
	}

	if(congratulations)
		delete congratulations;

	if(board)
		delete board;
}





bool MahJonggGL :: setRotScale(int val)
{
	if(val < 0)
		val = 1;

	switch(rotation_type)
	{
	case ROTATION_ARCBALL:
		ROT_SCALE = (float)val / 50.0f;
		break;
	default:
		ROT_SCALE = 0.0001f * (float) val;
		break;
	}

	return true;
}


void MahJonggGL :: mapToSphere( int x, int y, float result[3] )
{
	// scale x and y to within [-1. 1]
	float newx =        2.0f * (float) x / (float) window_size_x - 1.0f;
	float newy = 1.0f - 2.0f * (float) y / (float) window_size_y;

	// map to sphere
	float length = newx * newx + newy * newy;
	if(length == 0)
		return;

	if (length < 1.0f)
	{
		//within unit circle: project to unit sphere
		result[0] = newx;
		result[1] = newy;
		result[2] = sqrt(1.0f - length);
	}
	else
	{
		//outside unit circle: project to unit sphere cross section
		float norm = 1.0f / sqrt(length);
		result[0] = newx / norm;
		result[1] = newy / norm;
		result[2] = 0.0f;
	}
}

void MahJonggGL::enableGL(bool enable)
{
	enabled = enable;
}

// setup the camera and the looking direction
void MahJonggGL :: setupCamera()
{
	if(!enabled)
		return;

	glLoadIdentity();

	float angle = acos(cosangle);

	switch(rotation_type)
	{
	case ROTATION_ARCBALL:

		glTranslatef  (0.0f, 0.0f, -CAMERA_DISTANCE);
		glRotatef     (ROT_SCALE*angle * 360.0f / PI, axis[0], axis[1], axis[2]);
		glMultMatrixf (transform);
		glTranslatef  (-board_width * TILE_WIDTH * 0.5f, -board_height * TILE_HEIGHT * 0.5f, 0.0f);
		break;

	default:
		gluLookAt(	cam_x,
		           cam_y,
		           cam_z,
		           camera_target_x,
		           camera_target_y,
		           camera_target_z,
		           0.0f, 0.0f, 1.0f);
		break;
	}
}

// Create the display lists of the tiles and load the textures
bool MahJonggGL :: createTileGLs(const char *tileset_graphics_filename)
{
	int x;
	char path[300];
	char string[300];
	char absfilepath[300];

	memset(path, '\0', 300);

	FILE *file = fopen(tileset_graphics_filename, "r");
	if(!file)
		return false;

	for(x = strlen(tileset_graphics_filename)-1; x > 0; x--)
	{
		if((tileset_graphics_filename[x] == '/') || (tileset_graphics_filename[x] == '\\'))
		{
			break;
		}
	}
	strncat(path, tileset_graphics_filename, x+1);

	for(x = 0; x < NUM_TEXTURES; x++)
	{
		fgets(string, 100, file);

		string[strlen(string)-1] = '\0';
		memset(absfilepath, '\0', 300);

		strcat(absfilepath, path);
		strcat(absfilepath, string);

		if(tiles[x])
			delete tiles[x];

		tiles[x] = new TileGL(6, 128, absfilepath);
		tiles[x]->initializeTile(false);
	}

	fclose(file);

	memset(path, '\0', 300);
#ifdef LINUX

	strcpy(path, GAMEDATA_BASE_PATH);
	strcat(path, "/");
#endif

	strcat(path, "gra/shadow.bmp");

	shadow_tile = new TileGL(6, 32, path);
	shadow_tile->initializeTile(true);
	return true;
}


// initialize OpenGL
void MahJonggGL :: initializeGL()
{
	if(!enabled)
		return;
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);

	glEnable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_CULL_FACE);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_CULL_FACE); // face culling: draw only visible faces
	glCullFace(GL_BACK);


	// check whether the system supports stencil buffers. If not, disable the stencil test
	int num;
	glGetIntegerv(GL_STENCIL_BITS, &num);
	if(num > 0)
	{
		stencil = true;
	}
	else
		stencil = false;

	glShadeModel(GL_SMOOTH);

	char path[300];
	memset(path, '\0', 300);
#ifdef LINUX

	strcpy(path, GAMEDATA_BASE_PATH);
	strcat(path, "/");
#endif

	char numpath[350];
	memset(numpath, '\0', 350);
	strcpy(numpath, path);
	strcat(numpath, "gra");
	CreateNumberDisplayLists(numpath);

	char layoutpath[350];
	memset(layoutpath, '\0', 350);
	strcpy(layoutpath, path);
	char tilesetpath[350];
	memset(tilesetpath, '\0', 350);
	strcpy(tilesetpath, path);
	char backgroundpath[350];
	memset(backgroundpath, '\0', 350);
	strcpy(backgroundpath, path);

	strcat(layoutpath, "layouts/turtle.layout");
	strcat(tilesetpath, "tilesets/default.tileset");
	strcat(backgroundpath, "backgrounds/papyrus.background.bmp");

	makeBoard(0, layoutpath, tilesetpath, backgroundpath);
}


// generate a solvable board and assign the openGL tiles to it
bool MahJonggGL::makeBoard(int random_seed, const char *layout, const char *tileset, const char *background)
{
	int x,y,z;
	int flowercount = 0;
	int seasoncount = 0;
	int tile_type;
	Tile *tile;

	int securitycount = 64;

	if(!board)
		board = new MahJonggBoard();
	while(!board->generateSolvableBoard(random_seed, layout))
	{
		securitycount--;
		if(securitycount == 0)
			return false;
	}

	createBackground(background);

	board->getBoardGeometry(&board_width, &board_height, &board_depth);

	camera_target_x = board_width * TILE_WIDTH * 0.5f;
	camera_target_y = board_height * TILE_HEIGHT * 0.5f;
	camera_target_z = 0;

	cam_x = CAMERA_DISTANCE * (float)cos(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_x;
	cam_y = CAMERA_DISTANCE * (float)sin(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_y;
	cam_z = CAMERA_DISTANCE * (float)sin(angle_z);

	createTileGLs(tileset);

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				tile = board->getTile(x,y,z);
				if((tile->getState() != TILE_UNMARKED) && (tile->getState() != TILE_MARKED))
					continue;

				tile_type = tile->getType();

				// flower and season tiles are of the same type, but have different textures:
				if(tile_type == 35)
				{
					if(seasoncount == 4)
						seasoncount = 3;
					tile->setGLTile(tiles[38+seasoncount]);
					seasoncount++;
				}
				else if(tile_type == 34)
				{
					if(flowercount == 4)
						flowercount = 3;
					tile->setGLTile(tiles[34+flowercount]);
					flowercount++;
				}
				else
				{
					tile->setGLTile(tiles[tile_type]);
				}
			}
	return true;
}


void MahJonggGL :: resizeGL(GLsizei width, GLsizei height)
{
	if(!enabled)
		return;
	glLoadIdentity();

	window_size_x = width;
	window_size_y = height;

	if(window_size_y == 0)
	{
		height = window_size_y = 1;
	}

	glMatrixMode(GL_PROJECTION);
	glViewport(	(GLint) 0,
	            (GLint) 0,
	            (GLsizei) width,
	            (GLsizei) height);

	glLoadIdentity();

	paintGL();
}


// this method paints the scene and casts the shadows
void MahJonggGL :: paintGL()
{
	if(!enabled)
		return;

	int o;
	div_t division;

	float shadow_matrix[4][4];
	float plane_points[3][3];

	glMatrixMode(GL_PROJECTION);

	glLoadIdentity();
	gluPerspective(45.0f, (GLfloat)window_size_x / (GLfloat)window_size_y, VIEWNEARDIST, VIEWDEPTH);

	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, Omni0Pos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, Omni0Color);
	glLightfv(GL_LIGHT0, GL_AMBIENT, Omni0Amb);
	glEnable(GL_LIGHT0);


	glMatrixMode(GL_MODELVIEW);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	setupCamera();

	if(!board)
		return;
	if(!board->isInitialized())
		return;

	if(background_texture != 0)
	{
		glDisable(GL_DEPTH_TEST);
		glPushMatrix();

		glLoadIdentity();

		drawBackground();

		glPopMatrix();
		glEnable(GL_DEPTH_TEST);
	}

	// draw the scene
	for(o=0; o<board_depth; o++)
	{
		RenderSelectionScene(o, false);
	}

	// draw the shadows of the tiles
	if(shadow)
	{
		if(stencil)
		{
			glEnable(GL_STENCIL_TEST);

			glStencilFunc(GL_EQUAL, 0, 1);
			glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
			glClear(GL_STENCIL_BUFFER_BIT);
		}
		for(o=0; o<board_depth-1; o++)
		{
			glPushAttrib(GL_LIGHTING_BIT);
			glDisable(GL_LIGHTING);
			glPushMatrix();

			plane_points[1][0] = 0.0f;
			plane_points[1][1] = 0.0f;
			plane_points[1][2] = TILE_DEPTH*o + 0.1f;

			plane_points[0][0] = TILE_WIDTH;
			plane_points[0][1] = 0.0f;
			plane_points[0][2] = TILE_DEPTH*o + 0.1f;

			plane_points[2][0] = 0.0f;
			plane_points[2][1] = TILE_HEIGHT;
			plane_points[2][2] = TILE_DEPTH*o + 0.1f;

			CreateShadowMatrix(plane_points, Omni0Pos, shadow_matrix);

			glMultMatrixf((GLfloat *)shadow_matrix);

			RenderSelectionScene(o, true);

			glPopMatrix();
			glPopAttrib();
		}
		glDisable(GL_STENCIL_TEST);
	}

	// Draw the number of removable pairs
	division = div(board->calcNumRemovablePairs(), 10);
	glPushMatrix();
	glLoadIdentity();

	// if the number of removable pairs is zero, the game ends
	if((division.quot == 0) && (division.rem == 0))
	{
		char string[300];
		memset(string, '\0', 300);
#ifdef LINUX

		strcpy(string, GAMEDATA_BASE_PATH);
		strcat(string, "/");
#endif

		strcat(string, "gra/");

		DisplayCongratulations(string);
	}
	else // display the number of remaining pairs and tiles
	{
		glTranslatef(-5.2f, 3.0f, -10.0f);
		glPushAttrib(GL_LIGHTING_BIT);
		glPushAttrib(GL_DEPTH_BUFFER_BIT);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);

		glBindTexture(GL_TEXTURE_2D, number_texnum[division.quot]);
		glCallList(division.quot + 100);

		glTranslatef(NUMBER_WIDTH, 0.0f, 0.0f);

		glBindTexture(GL_TEXTURE_2D, number_texnum[division.rem]);
		glCallList(division.rem + 100);

		glTranslatef(2*NUMBER_WIDTH, 0.0f, 0.0f);

		division = div(board->getNumTilesOnBoard(), 100);
		glBindTexture(GL_TEXTURE_2D, number_texnum[division.quot]);
		glCallList(division.quot + 100);

		division = div(division.rem, 10);
		glTranslatef(NUMBER_WIDTH, 0.0f, 0.0f);
		glBindTexture(GL_TEXTURE_2D, number_texnum[division.quot]);
		glCallList(division.quot + 100);

		glTranslatef(NUMBER_WIDTH, 0.0f, 0.0f);
		glBindTexture(GL_TEXTURE_2D, number_texnum[division.rem]);
		glCallList(division.rem + 100);

		glPopAttrib();
		glPopAttrib();
	}
	glPopMatrix();
}


bool MahJonggGL::setRotationType(int type)
{
	switch(type)
	{
	case ROTATION_DEFAULT:
		break;
	case ROTATION_ARCBALL:
		break;
	default:
		return false;
	}

	rotation_type = type;
	return true;
}

bool MahJonggGL :: isMouseRightPressed()
{
	return mb_r;
}


// render the designated layer. If shadow is true, only draw shadow tiles
void MahJonggGL :: RenderSelectionScene(int layer, bool shadow)
{
	int m,n,o;
	int listname;
	int coord;
	int special;

	o = layer;

	for(m=0; m < board_width; m++)
		for(n=0; n < board_height; n++)
		{
			if((board->getTile(m,n,o)->getState() == TILE_LOCKED) || (board->getTile(m,n,o)->getState() == TILE_EMPTY))
			{
				continue;
			}
			coord = board->convertCoordinates(m,n,o);

			glPushMatrix();
			glPushName(coord);

			board->getTile(m,n,o)->getSpecial(&special);

			if(special == -1)
				glTranslatef((float)((float)m*(TILE_WIDTH)), (float)((float)n*(TILE_HEIGHT)), (float)((float)o*TILE_DEPTH));
			else if(special == 2)
				glTranslatef((float)((float)m*(TILE_WIDTH)), (float)((float)n*(TILE_HEIGHT)) + 0.5f*TILE_HEIGHT, (float)((float)o*TILE_DEPTH));
			else if(special == 4)
				glTranslatef((float)((float)m*(TILE_WIDTH)) + 0.5f*TILE_WIDTH, (float)((float)n*(TILE_HEIGHT)) + 0.5f*TILE_HEIGHT, (float)((float)o*TILE_DEPTH));
			else
			{
				glPopName();
				glPopMatrix();
				continue;
			}

			listname = board->getDisplayList(m,n,o);
			if(listname != 0)
			{
				if(!shadow)
				{
					if(board->getTextureName(m,n,o) == shadow_tile->getTextureName())
					{
						glBindTexture(GL_TEXTURE_2D, board->getTextureName(m,n,o));	// use the texture that was assigned to the tile
					}
					glBindTexture(GL_TEXTURE_2D, board->getTextureName(m,n,o));	// use the texture that was assigned to the tile
					glCallList(listname);	// draw the tile (by calling the corresponding tile_type as display list id)
				}
				else
				{
					glBindTexture(GL_TEXTURE_2D, shadow_tile->getTextureName());
					glCallList(shadow_tile->getDisplayListUnmarked());
				}
			}

			glPopName();
			glPopMatrix();
		}

	glFlush();
}

void MahJonggGL :: mouseRotationDefault(int x, int y)
{
	angle_x += ((float)x - (float)mouse_x)*ROT_SCALE;
	angle_z += ((float)y - (float)mouse_y)*ROT_SCALE;

	while(angle_x >= (2*PI))
		angle_x -= 2*PI;

	while(angle_x < 0)
		angle_x += 2*PI;

	if(angle_z >= (0.45f*PI))
		angle_z = 0.45f*PI;

	while(angle_z < 0)
		angle_z = 0.0f;

	cam_x = CAMERA_DISTANCE * (float)cos(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_x;
	cam_y = CAMERA_DISTANCE * (float)sin(angle_x) * (float)cos(angle_z) + camera_target_z + camera_target_y;
	cam_z = CAMERA_DISTANCE * (float)sin(angle_z);

	mouse_x = x;
	mouse_y = y;
}


void MahJonggGL :: mouseRotationArcBall(int x, int y)
{
	float drag_here[3];
	mapToSphere (x, y, drag_here);
	cross (drag_start, drag_here, axis);
	cosangle = dot (drag_start, drag_here);
}

void MahJonggGL :: mouseMoveEvent(int x, int y, int button)
{
	if(!enabled)
		return;
	if(!board)
		return;

	if(mb_r)	// if the right mouse button is pressed, rotate the camera
	{
		switch(rotation_type)
		{
		case ROTATION_ARCBALL:
			mouseRotationArcBall(x,y);
			break;
		default:
			mouseRotationDefault(x,y);
			break;
		}
	}
}


// handle mouse press events, the left mouse button selects tiles
void MahJonggGL :: mousePressEvent(int x, int y, int button)
{
	if(!enabled)
		return;
	if(!board)
		return;

	GLuint select_buffer[PICKING_BUFFER];
	GLuint *sbuf;
	GLint viewport[4];
	GLint hits;
	GLuint name = 0, num_names = 0, selected_name = 0;
	float depth = 0.0f, selected_depth = 0.0f;
	bool change = false;

	int m,n, o;

	for(m = 0; m < PICKING_BUFFER; m++)
		select_buffer[m] = 0;

	if(button == MOUSE_RIGHT) // right clicking and holding enables the rotation of the view
	{
		mb_r = true;
		switch(rotation_type)
		{
		case ROTATION_DEFAULT:
			mouse_x = x;
			mouse_y = y;
			break;
		case ROTATION_ARCBALL:
			mapToSphere (x, y, drag_start);
			break;
		}

	}
	else if(button == MOUSE_LEFT) // left clicking enters selection mode
	{
		switch(rotation_type)
		{
		case ROTATION_DEFAULT:
			mouse_x = x;
			mouse_y = y;
			break;

		case ROTATION_ARCBALL:
			mouse_x = x;
			mouse_y = y;
			break;
		}
		glGetIntegerv(GL_VIEWPORT, viewport); // get the viewport
		glSelectBuffer(PICKING_BUFFER, select_buffer); // initialize the selection buffer
		glRenderMode(GL_SELECT);	// enter the selection mode
		glInitNames();
		glPushName(0);

		glPushMatrix();

		// restrict drawing to a small region around the mouse pointer
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPickMatrix((GLdouble)mouse_x, (GLdouble)(viewport[3] - mouse_y), 5.0f, 5.0f, viewport);
		gluPerspective(45.0f, (float)window_size_x / (float)window_size_y, VIEWNEARDIST, VIEWDEPTH);

		glMatrixMode(GL_MODELVIEW);

		setupCamera();

		// draw the scene without shadows
		for(o = 0; o < board_depth; o++)
			RenderSelectionScene(o, false);

		glPopMatrix();

		hits = glRenderMode(GL_RENDER); // retrieve hits

		// now the selection buffer is processed
		if(hits != 0)
		{
			sbuf = (GLuint *)select_buffer;

			for(m = 0; m < hits; m++)
			{
				num_names = *sbuf;
				sbuf++;
				depth = *sbuf;

				// select the tile that is nearest to the viewer
				if((selected_depth > depth) || (selected_depth == 0.0f))
				{
					selected_depth = depth;
					change = true;
				}

				sbuf++;
				sbuf++;
				for(n = 0; n < num_names; n++)
				{
					sbuf++;
					name = *sbuf;
					if(change)
						selected_name = name;
					change = false;
				}
			}

			select(selected_name);
		}
	}
}


// this is a quite slow implementation of a list search engine --
// -- perhaps implement it as a bubble-sort or quicksort engine???
void MahJonggGL :: select(int name)
{
	if(!enabled)
		return;
	int x,y,z;
	int coord;

	for(x = 0; x < board_width; x++)
		for(y = 0; y < board_height; y++)
			for(z = 0; z < board_depth; z++)
			{
				coord = board->convertCoordinates(x,y,z);
				if(name == coord)
					board->selectTile(x,y,z);
			}
}


void MahJonggGL :: mouseReleaseEvent(int x, int y, int button)
{
	if(!enabled)
		return;
	if(!board)
		return;

	if(button == MOUSE_RIGHT)
	{
		mb_r = false;

		switch(rotation_type)
		{
		case ROTATION_DEFAULT:
			break;

		case ROTATION_ARCBALL:
			// save the current rotation
			glMatrixMode (GL_MODELVIEW);
			glPushMatrix ();
			glLoadIdentity ();
			glRotatef      (ROT_SCALE*acos (cosangle) * 360.0f / PI, axis[0], axis[1], axis[2]);
			glMultMatrixf  (transform);
			glGetFloatv    (GL_MODELVIEW_MATRIX, transform);
			glPopMatrix ();

			cosangle = 1.0f;
			break;
		}
	}
	else
	{
		pick = false;
	}
}


// this creates the flattening shadow matrix
void MahJonggGL :: CreateShadowMatrix(GLfloat points[3][3], GLfloat light[4], GLfloat matrix[4][4])
{
	GLfloat plane[4];
	GLfloat dotProduct;
	GLfloat len;

	GLfloat vec1[3], vec2[3], norm[3];

	// the points array represent the plane onto which the shadow will be projected.
	// vec* are direction vectors from point[1] to points[0] and [2]
	vec1[0] = points[0][0] - points[1][0];
	vec1[1] = points[0][1] - points[1][1];
	vec1[2] = points[0][2] - points[1][2];

	vec2[0] = points[2][0] - points[1][0];
	vec2[1] = points[2][1] - points[1][1];
	vec2[2] = points[2][2] - points[1][2];

	// calculate the normal of the plane by computing the cross product of the direction vectors
	norm[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
	norm[1] = -vec1[0] * vec2[2] + vec1[2] * vec2[0];
	norm[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

	len = (float)sqrt(norm[0] * norm[0] + norm[1] * norm[1] + norm[2] * norm[2]);
	if(len == 0)
		return;

	// normalize the normal  and set it as the plane normal
	plane[0] = norm[0] / len;
	plane[1] = norm[1] / len;
	plane[2] = norm[2] / len;

	// the last coefficient of the plane represents the distance of the plane to the origin
	plane[3] = -(plane[0] * points[0][0] + plane[1] * points[0][1] + plane[2] * points[0][2]);

	dotProduct = plane[0] * light[0] + plane[1] * light[1] + plane[2] * light[2] + plane[3] * light[3];

	// create a flattening matrix out of the light position and the shadow projection plane
	matrix[0][0] = dotProduct - light[0] * plane[0];
	matrix[1][0] = - light[0] * plane[1];
	matrix[2][0] = - light[0] * plane[2];
	matrix[3][0] = - light[0] * plane[3];

	matrix[0][1] = - light[1] * plane[0];
	matrix[1][1] = dotProduct - light[1] * plane[1];
	matrix[2][1] = - light[1] * plane[2];
	matrix[3][1] = - light[1] * plane[3];

	matrix[0][2] = - light[2] * plane[0];
	matrix[1][2] = - light[2] * plane[1];
	matrix[2][2] = dotProduct - light[2] * plane[2];
	matrix[3][2] = - light[2] * plane[3];

	matrix[0][3] = - light[3] * plane[0];
	matrix[1][3] = - light[3] * plane[1];
	matrix[2][3] = - light[3] * plane[2];
	matrix[3][3] = dotProduct - light[3] * plane[3];
}


// create display lists for the numbers
void MahJonggGL :: CreateNumberDisplayLists(char *directory)
{
	if(!directory)
		return;
	if(strlen(directory) == 0)
		return;
	// the color of the number material

	unsigned int n;
	unsigned int texname;

	char string[300];

	for(n=0; n < 10; n++)
	{
		//		if(number_bmp[n]) delete number_bmp[n];
		number_bmp[n] = new CBitmapFile();
	}

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/0.bmp");
	number_bmp[0]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/1.bmp");
	number_bmp[1]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/2.bmp");
	number_bmp[2]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/3.bmp");
	number_bmp[3]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/4.bmp");
	number_bmp[4]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/5.bmp");
	number_bmp[5]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/6.bmp");
	number_bmp[6]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/7.bmp");
	number_bmp[7]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/8.bmp");
	number_bmp[8]->LoadBMP(string);

	memset(string, '\0', 300);
	strcpy(string, directory);
	strcat(string, "/9.bmp");
	number_bmp[9]->LoadBMP(string);

	for(n = 0; n < 10; n++)
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glGenTextures(1, &texname);
		number_texnum[n] = texname;

		glBindTexture(GL_TEXTURE_2D, texname);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 64, 64, 0, GL_RGB, GL_UNSIGNED_BYTE, number_bmp[n]->image->data);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glNewList(n + 100, GL_COMPILE);
		glColor4f(1.0f, 1.0f, 1.0f, 0.7f);

		glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, 1.0f);

		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(NUMBER_WIDTH, 0.0f, 0.0f);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(NUMBER_WIDTH, NUMBER_HEIGHT, 0.0f);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(0.0f, NUMBER_HEIGHT, 0.0f);

		glEnd();
		glEndList();
	}
}


// endgame screen
void MahJonggGL :: DisplayCongratulations(char *directory)
{
	if(!directory)
		return;
	if(strlen(directory) == 0)
		return;

	if(congratulations)
		delete congratulations;
	char string[300];
	memset(string, '\0', 300);

	strcpy(string, directory);
	strcat(string, "/");

	congratulations = new CBitmapFile();
	if(board->getNumTilesOnBoard() == 0)
		strcat(string, "congratulations.bmp");

	else
		strcat(string, "no_more_moves.bmp");

	congratulations->LoadBMP(string);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(1, &congratulations_texnum);

	glBindTexture(GL_TEXTURE_2D, congratulations_texnum);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 128, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, congratulations->image->data);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glPushAttrib(GL_LIGHTING_BIT);
	glPushAttrib(GL_DEPTH_BUFFER_BIT);

	glDisable(GL_DEPTH_TEST);

	glTranslatef(-50.0f, -20.0f, -100.0f);
	glBegin(GL_QUADS);
	glNormal3f(0.0f, 0.0f, 1.0f);

	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f);
	glVertex3f(CONGRATULATIONS_WIDTH, 0.0f, 0.0f);
	glTexCoord2f(1.0f, 1.0f);
	glVertex3f(CONGRATULATIONS_WIDTH, CONGRATULATIONS_HEIGHT, 0.0f);
	glTexCoord2f(0.0f, 1.0f);
	glVertex3f(0.0f, CONGRATULATIONS_WIDTH, 0.0f);

	glEnd();

	glPopAttrib();
	glPopAttrib();
}



/*!
    \fn MahJonggGL::createBackground(const char *tex_filename)
 */
unsigned int MahJonggGL::createBackground(const char *tex_filename)
{
	char string[300];
	memset(string, '\0', 300);

	strcat(string, tex_filename);

	if(bg_tex)
		delete bg_tex;
	bg_tex = new CBitmapFile();
	bg_tex->LoadBMP(string);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(1, &background_texture);
	glBindTexture(GL_TEXTURE_2D, background_texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 128, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, bg_tex->image->data);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	return background_texture;
}


/*!
    \fn MahJonggGL::drawBackground()
 */
void MahJonggGL::drawBackground()
{
	float x1,y1;
	float x2,y2;
	float width = (float)window_size_x;
	float height = (float)window_size_y;

	x1 = 0.5f * width / height *( - board_width * TILE_WIDTH * 0.5f);
	y1 = 0.5f * width / height *( - board_height * TILE_HEIGHT * 0.5f);
	x2 = 0.5f * width / height *( board_width * TILE_WIDTH * 0.5f);
	y2 = 0.5f * width / height *( board_height * TILE_HEIGHT * 0.5f);

	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D, background_texture);

	glBegin(GL_QUADS);

	glColor4f(1.0f,1.0f,1.0f,1.0f);

	glTranslatef(0.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(x1, y1, -10.0f);
	glTexCoord2f(10.0f, 0.0f);
	glVertex3f(x2, y1, -10.0f);
	glTexCoord2f(10.0f, 10.0f);
	glVertex3f(x2, y2, -10.0f);
	glTexCoord2f(0.0f, 10.0f);
	glVertex3f(x1, y2, -10.0f);

	glEnd();
	glPopAttrib();
	glPopAttrib();
}


